#include "entity_health.hpp"
#include "unit.hpp"
#include "utils/godot_macros.hpp"

void EntityHealth::_bind_methods() {
#define CLASSNAME EntityHealth
    GDFUNCTION_ARGS(damaged_by, "amount", "source");
    GDFUNCTION_ARGS(healed_by, "amount", "source");
    GDPROPERTY(injury_max, gd::Variant::INT);
    GDPROPERTY(wounds_max, gd::Variant::INT);

    GDSIGNAL("damage",
             gd::PropertyInfo(gd::Variant::OBJECT, "health_entity", gd::PROPERTY_HINT_NODE_TYPE, "EntityHealth"),
             gd::PropertyInfo(gd::Variant::INT, "change"),
             gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
    GDSIGNAL("heal",
             gd::PropertyInfo(gd::Variant::OBJECT, "health_entity", gd::PROPERTY_HINT_NODE_TYPE, "EntityHealth"),
             gd::PropertyInfo(gd::Variant::INT, "change"),
             gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
    GDSIGNAL("health_changed",
             gd::PropertyInfo(gd::Variant::OBJECT, "health_entity", gd::PROPERTY_HINT_NODE_TYPE, "EntityHealth"),
             gd::PropertyInfo(gd::Variant::INT, "change"));
    GDSIGNAL("death", gd::PropertyInfo(gd::Variant::OBJECT, "source", gd::PROPERTY_HINT_NODE_TYPE, "Unit"));
}

void EntityHealth::_enter_tree() {
    this->injury_current = this->injury_max;
    this->wounds_current = this->wounds_max;
}

void EntityHealth::damaged_by(int amount, Unit *source) {
    // do not take damage when already dead
    if(this->injury_current <= 0)
        return;
    amount = gd::Math::abs(amount);
    this->wounds_current -= amount;
    this->injury_current -= gd::Math::max(1, int(float(amount) * this->get_wounds_damage_factor()));

    this->emit_signal("damage", this, amount, source);
    this->emit_signal("health_changed", this, -amount);
    if(this->injury_current <= 0)
        this->emit_signal("death", source);
}

void EntityHealth::healed_by(int amount, Unit *source) {
    // reviving takes a different action
    if(this->injury_current <= 0)
        return;
    amount = gd::Math::abs(amount);
    this->injury_current = gd::Math::min(this->injury_max, this->injury_current + amount);
    this->emit_signal("heal", this, amount, source);
    this->emit_signal("health_changed", this, amount);
}

float EntityHealth::get_wounds_damage_factor() const {
    return float(this->injury_current) / float(this->injury_max);
}

bool EntityHealth::get_is_dead() const {
    return this->injury_current <= 0;
}

void EntityHealth::set_injury_max(int max_health) {
    this->injury_max = max_health;
}

int EntityHealth::get_injury_max() const {
    return this->injury_max;
}

int EntityHealth::get_injury_current() const {
    return this->injury_current;
}

void EntityHealth::set_wounds_max(int max_health) {
    this->wounds_max = max_health;
}

int EntityHealth::get_wounds_max() const {
    return this->wounds_max;
}

int EntityHealth::get_wounds_current() const {
    return this->wounds_current;
}
